import * as Mqtt from 'mqtt'
import * as Mongo from 'mongodb'
import * as Events from 'events'
import * as Queue from './queue'
import * as MqttMessage from './mqtt_message'


let console = require('chalk-console')


const MQTT_BROKER_URL : String = 'mqtt://localhost'
const MONGO_DATABASE_URL : String = 'mongodb://localhost/test'
const COLLECTION : String = 'mqttMongo'


function collect(val: string, memo: string[]) {
  memo.push(val);
  return memo;
}

import options = require('commander')
options
  .version('0.0.1')
  .option('-b, --broker [broker url]', 'Broker URL', MQTT_BROKER_URL)
  .option('-m, --database [database url]', 'MongoDB Database URL', MONGO_DATABASE_URL)
  .option('-c, --collection [mongodb collection]', 'Collection in MongoDB Database', COLLECTION)
  .option('-t, --topic [topic to subscribe]', 'Topic to subscribe, can appear multiple times', collect, [])
  .option('-e, --encapsulate', 'store current timestamp, topic and payload in document', false)
  .option('-p, --parsePayload', 'parse payload when encapsulating (otherwise always)', false)
  .parse(process.argv)


class MqttMongo extends Events.EventEmitter {

  private options : any

  constructor(options : any) {
    super()
    this.options = options
    this.on('reconnectDatabase', this.connectToDatabase)
  }

  private dbHandle : Mongo.Db;
  private dbReady : boolean = false;
  private queue : Queue.Queue<MqttMessage.MqttMessage> = new Queue.Queue<MqttMessage.MqttMessage>()
  private mqttClient : Mqtt.Client
  private heartbeatTimer : NodeJS.Timer
  private uptime : number = 0

  connectToDatabase() {
    console.info("About to connect to database")
    Mongo.MongoClient.connect(this.options.database) 
      .then(
        (tmpDbHandle: Mongo.Db) => { 
          this.dbHandle = tmpDbHandle
          this.dbReady = true; 
          console.info("Database connected")
          this.dbHandle.on('reconnectFailed', (err : any) => { console.warn(`Error on database ${err}`) })
          this.dbHandle.on('reconnect', () => { 
            console.info("Reconnect on database") 
            this.dbReady = true
            this.queue.knock()
          })
          this.dbHandle.on('timeout', () => { console.warn("Timeout on database") })
          this.dbHandle.on('close', () => { 
            console.info("Close on database") 
            this.dbReady = false
          })
        },
        (err: String) => {
          console.error(`Unable to connect to database: ${err}`)
          this.dbReady = false
        }
    )
  }

  connectToBroker() {
    this.queue.on('data', () => {
      if (this.dbReady || true) { 
        while (! this.queue.isEmpty()) {
          let msg : MqttMessage.MqttMessage = this.queue.deq()
          console.info(`Something in the queue: ${JSON.stringify(msg)}`)
          let coll = this.dbHandle.collection(this.options.collection)
          coll.insertOne(msg.getMessage())
            .then(
              (r) => { 
                console.success(`Successfully inserted into database ${JSON.stringify(msg.getMessage())}`)
              },
              (err) => { 
                console.error(`Error when trying to insert into database ${err}`)
                if (! this.dbReady) {
                  console.info("Error occured while database connection is lost, re-enqueue msg.")
                  this.queue.reenq(msg)
                  this.emit('reconnectDatabase')
                } else {
                  console.error(`Message ${JSON.stringify(msg.getMessage())} is lost`)
                }
              }
            )
        }
      } else {
        // console.info("Database currently not available, not reading from stream")
      }
    })
  
    this.mqttClient = Mqtt.connect(this.options.broker)
    this.mqttClient.on('offline', () => { console.warn("MQTT client is offline") })
    this.mqttClient.on('reconnect', () => { console.warn("MQTT client is reconnecting") })
    this.mqttClient.on('close', () => { console.warn("MQTT connection closed") })

    this.mqttClient.on('connect', () => {
      console.info("MQTT broker connected")
      this.options.topic.forEach((topic: string) => {
        this.mqttClient.subscribe(topic)
        console.info(`Subscribed to ${topic}`)
      })
      this.mqttClient.subscribe('MqttMongo/Command')
      this.mqttClient.publish('MqttMongo/Status', 'hello, started up')
    })

    let msgCnt : number = 0
    this.mqttClient.on('message', (topic : string, messageBuf : Buffer) => {
      msgCnt++;
      let message = messageBuf.toString('UTF-8')
      console.info(`Message received ${msgCnt}, topic ${topic}, payload ${message}`)

      if (topic == "MqttMongo/Command" && message == "shutdown") {
        this.shutdown()
      } else {
        try {
          this.queue.enq(new MqttMessage.MqttMessage(topic, message, this.options.encapsulate, this.options.parsePayload))
        } catch (e) {
          console.error(`Error while parsing Mqtt message, topic '${topic}', message '${message}' , ${e.toString()}`)
        }
      }
    })
  }

  setupHeartbeat() {
    this.heartbeatTimer = setInterval(() => {
      this.uptime++
      this.mqttClient.publish('MqttMongo/Status', `{'Uptime': ${this.uptime}}`)
      if (! this.dbReady) {
        this.emit("reconnectDatabase")
      }
    }, 1000)
    console.info("Heartbeat timer started")
  }

  shutdown() {
    console.info("Shutting down MqttMongo")
    clearInterval(this.heartbeatTimer)
    this.mqttClient.end()
    this.dbHandle.close()
  }
}

let mqttMongo : MqttMongo = new MqttMongo(options)
mqttMongo.connectToDatabase()
mqttMongo.connectToBroker()
mqttMongo.setupHeartbeat()

console.info("MqttMongo started")